package sitzplangenerator;

import java.io.*;
import java.util.LinkedList;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.TableModel;

/**
 * @author B. A. Bartsch
 */
public final class Model implements TableModel {

    public final int ROOM_DX = 75;
    public final int ROOM_DY = 75;
    public final int TABLE_W = 40;
    public final int TABLE_H = 40;

    public static final int TABLE_NO = 0;
    public static final int TABLE_YES = 1;
    public static final int TABLE_BLOCKED = 2;

    public static final int PRINT_VORNAME = 1;
    public static final int PRINT_NACHNAME = 2;
    public static final int PRINT_VORNAME_NACHNAME = 3;
    public static final int PRINT_NACHNAME_VORNAME = 4;

    private static final String HEADER = "T A F E L";
    private static final String HEADER_NUMMER = "Nummer";
    private static final String HEADER_NACHNAME = "Nachname";
    private static final String HEADER_VORNAME = "Vorname";

    private final LinkedList<ModelChangeListener> listenersAll = new LinkedList<ModelChangeListener>();

    // DATA
    private boolean viewBehind = true;
    private boolean showBorder = true;
    private boolean showTableThicker = false;
    private int printState = PRINT_VORNAME_NACHNAME;
    private File lastOpenedDirectory = null;

    // ### FORBIDDEN ###
    private boolean[][] forbiddenPair;
    private String forbiddenText;
    private int forbiddenCount;

    // ### ALLOWED ###
    private boolean[][] allowedPair;
    private String allowedText;
    private int allowedCount;

    // ### DATA NAMEN ###
    private final LinkedList<TableModelListener> listenersNames = new LinkedList<TableModelListener>();

    private int MAX_PERSONS;
    private String[] namesNachnameAt;
    private String[] namesVornameAt;

    // ### DATA TABLES ###
    private final int MAX_ROWS = 15;
    private final int MAX_COLS = 11;

    private String[][] roomVornameAt;
    private int[][] roomTableCodeAt;
    private String[][] roomNachnameAt;
    private int[][] roomPersonenNummerAt;

    // ### MOUSE ###
    private int mouseStartX = -1;
    private int mouseStartY = -1;
    private int mouseCurrentX = -1;
    private int mouseCurrentY = -1;

    public Model(int personen) {
        MAX_PERSONS = personen;
        //MAX_ROWS = 14;
        //MAX_COLS = 11;
        setClearAll();
    }

    // ### GETTER ###
    public String getHeader() {
        return HEADER;
    }

    // ### GENERELL ###
    public void setClearAll() {
        allowedSetClear();
        forbiddenSetClear();
        namesSetClear(MAX_PERSONS);
        roomSetClear();
        fireModelChanged();
    }

    public void setViewBehind(boolean b) {
        viewBehind = b;
        fireModelChanged();
    }

    public void setShowBorder(boolean b) {
        showBorder = b;
        fireModelChanged();
    }

    public void setShowTableThicker(boolean b) {
        showTableThicker = b;
        fireModelChanged();
    }

    public boolean getShowBorder() {
        return showBorder;
    }

    public boolean getShowTableThicker() {
        return showTableThicker;
    }

    public boolean getViewSchueler() {
        return viewBehind;
    }

    public int getPrintState() {
        return printState;
    }

    public void setPrintState(int state) {
        printState = state;
        fireModelChanged();
    }

    // ### FORBIDDEN ###
    public void forbiddenSetClear() {
        forbiddenPair = new boolean[MAX_PERSONS + 1][MAX_PERSONS + 1];
        forbiddenSetText("");
    }

    public int getForbiddenCount() {
        return forbiddenCount;
    }

    public String getForbiddenText() {
        return forbiddenText;
    }

    public void forbiddenSetText(String text) {
        forbiddenText = text;

        // Alle Paare sind zunächst erlaubt
        for (int i = 1; i <= MAX_PERSONS; i++) {
            for (int j = 1; j <= MAX_PERSONS; j++) {
                forbiddenPair[i][j] = false;
            }
        }

        // for each row
        String[] rowAt = text.split(Static.ASCII_CODE_10);
        for (String rowAt1 : rowAt) {
            try {
                String[] split = rowAt1.split(",");
                if (split.length < 2) {
                    continue;
                }
                for (int i1 = 0; i1 < split.length; i1++) {
                    for (int i2 = i1 + 1; i2 < split.length; i2++) {
                        int v1 = Integer.parseInt(split[i1]);
                        int v2 = Integer.parseInt(split[i2]);
                        if (v1 == v2) {
                            throw new Exception("v1==v2");
                        }
                        if ((v1 < 1) || (v1 > MAX_PERSONS)) {
                            throw new Exception("v1=" + v1);
                        }
                        if ((v2 < 1) || (v2 > MAX_PERSONS)) {
                            throw new Exception("v2=" + v2);
                        }
                        forbiddenPair[v1][v2] = true;
                        forbiddenPair[v2][v1] = true;
                    }
                }
            } catch (Exception ex) {
            }
        }

        // Zähle Paare
        forbiddenCount = 0;
        for (int i = 1; i <= MAX_PERSONS; i++) {
            for (int j = 1; j <= MAX_PERSONS; j++) {
                if (forbiddenPair[i][j]) {
                    forbiddenCount++;
                }
            }
        }
        forbiddenCount /= 2;

        fireModelChanged();
    }

    // ### ALLOWED ###
    public void allowedSetClear() {
        allowedPair = new boolean[MAX_PERSONS + 1][MAX_PERSONS + 1];
        allowedSetText("");
    }

    public int getAllowedCount() {
        return allowedCount;
    }

    public String getAllowedText() {
        return allowedText;
    }

    public void allowedSetText(String text) {
        allowedText = text;

        // reset
        for (int i = 1; i <= MAX_PERSONS; i++) {
            for (int j = 1; j <= MAX_PERSONS; j++) {
                allowedPair[i][j] = false;
            }
        }

        // for each row
        String[] rowAt = text.split(Static.ASCII_CODE_10);
        for (String rowAt1 : rowAt) {
            try {
                String[] split = rowAt1.split(",");
                if (split.length != 2) {
                    continue;
                }
                int v1 = Integer.parseInt(split[0]);
                int v2 = Integer.parseInt(split[1]);
                if (v1 == v2) {
                    throw new Exception("v1==v2");
                }
                if ((v1 < 1) || (v1 > MAX_PERSONS)) {
                    throw new Exception("v1=" + v1);
                }
                if ((v2 < 1) || (v2 > MAX_PERSONS)) {
                    throw new Exception("v2=" + v2);
                }
                allowedPair[v1][v2] = true;
                allowedPair[v2][v1] = true;
            } catch (Exception ex) {
            }
        }

        // Zähle Paare
        allowedCount = 0;
        for (int i = 1; i <= MAX_PERSONS; i++) {
            for (int j = 1; j <= MAX_PERSONS; j++) {
                if (allowedPair[i][j]) {
                    allowedCount++;
                }
            }
        }
        allowedCount /= 2;

        fireModelChanged();
    }

    // ### ROOM ###
    public boolean roomGetTableThickerActivated() {
        return showTableThicker;
    }

    public void roomSetClear() {
        roomTableCodeAt = new int[MAX_ROWS][MAX_COLS];
        roomVornameAt = new String[MAX_ROWS][MAX_COLS];
        roomNachnameAt = new String[MAX_ROWS][MAX_COLS];
        roomPersonenNummerAt = new int[MAX_ROWS][MAX_COLS];

        for (int r = 0; r < MAX_ROWS; r++) {
            for (int c = 0; c < MAX_COLS; c++) {
                roomTableCodeAt[r][c] = TABLE_NO;
                roomVornameAt[r][c] = "";
                roomNachnameAt[r][c] = "";
                roomPersonenNummerAt[r][c] = 0;
            }
        }

        fireModelChanged();
    }

    public int roomGetTableCodeAt(int row, int col) {
        return roomTableCodeAt[row][col];
    }

    public int roomGetTablePersonenNummerAt(int row, int col) {
        return roomPersonenNummerAt[row][col];
    }

    public String roomGetTableNachnameAt(int row, int col) {
        return roomNachnameAt[row][col];
    }

    public String roomGetTableVornameAt(int row, int col) {
        return roomVornameAt[row][col];
    }

    public void roomSetTableThicker(boolean b) {
        showTableThicker = b;
        fireModelChanged();
    }

    public void roomSetTableCodeAt(int r, int c, int tableCode) {
        roomTableCodeAt[r][c] = tableCode;
        fireModelChanged();
    }

    public void roomSetMouseStartLeftClick(int mouseX, int mouseY) {
        int[] temp = mouse_to_Grid(mouseX, mouseY);
        if (temp == null) {
            return;
        }

        int row = temp[0];
        int col = temp[1];

        if (roomTableCodeAt[row][col] == TABLE_NO) {
            roomTableCodeAt[row][col] = TABLE_YES;
        } else {
            if (roomTableCodeAt[row][col] == TABLE_YES) {
                roomTableCodeAt[row][col] = TABLE_BLOCKED;
            } else {
                roomTableCodeAt[row][col] = TABLE_NO;
            }
        }

        fireModelChanged();
    }

    public void roomSetMouseStartRightClick(int mouseX, int mouseY) {
        mouseStartX = -1;
        mouseStartY = -1;
        mouseCurrentX = -1;
        mouseCurrentY = -1;

        int[] temp = mouse_to_Grid(mouseX, mouseY);
        if (temp == null) {
            return;
        }

        mouseStartX = mouseX;
        mouseStartY = mouseY;
        mouseCurrentX = mouseX;
        mouseCurrentY = mouseY;

        fireModelChanged();
    }

    public void roomSetMouseCurrent(int mouseX, int mouseY) {
        mouseCurrentX = mouseX;
        mouseCurrentY = mouseY;

        fireModelChanged();
    }

    public void roomSetMouseReleasedRight(int mouseX, int mouseY) {
        int[] temp1 = mouse_to_Grid(mouseStartX, mouseStartY);
        int[] temp2 = mouse_to_Grid(mouseX, mouseY);

        mouseStartX = -1;
        mouseStartY = -1;
        mouseCurrentX = -1;
        mouseCurrentY = -1;

        if ((temp1 != null) && (temp2 != null)) {
            int r1 = temp1[0];
            int c1 = temp1[1];
            int r2 = temp2[0];
            int c2 = temp2[1];

            if (roomTableCodeAt[r1][c1] == TABLE_YES) {
                if (roomTableCodeAt[r2][c2] == TABLE_YES) {
                    roomTauscheTische(r1, c1, r2, c2);
                }
            }
        }

        fireModelChanged();
    }

    private int[] mouse_to_Grid(int mouseX, int mouseY) {
        int posX = mouseX - ROOM_DX;
        int posY = mouseY - ROOM_DY;
        if (posX < 0) {
            return null;
        }
        if (posY < 0) {
            return null;
        }
        int col = posX / TABLE_W;
        int row = posY / TABLE_H;

        // Punktspiegelung
        if (!viewBehind) {
            col = (MAX_COLS - 1) - col;
            row = (MAX_ROWS - 1) - row;
        }

        if (col < 0) {
            return null;
        }
        if (col >= MAX_COLS) {
            return null;
        }
        if (row <= 0) {
            return null;
        }
        if (row >= MAX_ROWS) {
            return null;
        }

        return new int[]{row, col};
    }

    public int roomGetColumns() {
        return MAX_COLS;
    }

    public int roomGetRows() {
        return MAX_ROWS;
    }

    public int roomGetDeltaX() {
        return ROOM_DX;
    }

    public int roomGetDeltaY() {
        return ROOM_DY;
    }

    public int roomGetTableWidth() {
        return TABLE_W;
    }

    public int roomGetTableHeight() {
        return TABLE_H;
    }

    public int roomGetMouseStartRightClickX() {
        return mouseStartX;
    }

    public int roomGetMouseStartRightClickY() {
        return mouseStartY;
    }

    public int roomGetMouseCurrentX() {
        return mouseCurrentX;
    }

    public int roomGetMouseCurrentY() {
        return mouseCurrentY;
    }

    public void roomSetTableRandomSeatingOrder() {

        // Lösche Namen/Personennummer
        for (int r = 0; r < MAX_ROWS; r++) {
            for (int c = 0; c < MAX_COLS; c++) {
                roomVornameAt[r][c] = "";
                roomNachnameAt[r][c] = "";
                roomPersonenNummerAt[r][c] = 0;
            }
        }

        // Verteile Namen/Personennummer auf die Tische
        for (int r = 0, i = 0; r < MAX_ROWS; r++) {
            for (int c = 0; c < MAX_COLS; c++) {
                if (roomTableCodeAt[r][c] == TABLE_YES) {
                    if (i < MAX_PERSONS) {
                        roomVornameAt[r][c] = namesVornameAt[i];
                        roomNachnameAt[r][c] = namesNachnameAt[i];
                        roomPersonenNummerAt[r][c] = i + 1;
                        i++;
                    }
                }
            }
        }

        // Zähle Tische
        int numberOfTables = 0;
        for (int r = 0; r < MAX_ROWS; r++) {
            for (int c = 0; c < MAX_COLS; c++) {
                if (roomTableCodeAt[r][c] == TABLE_YES) {
                    numberOfTables++;
                }
            }
        }

        // Sammeln der Tische
        int[][] list = new int[numberOfTables][2];
        for (int r = 0, i = 0; r < MAX_ROWS; r++) {
            for (int c = 0; c < MAX_COLS; c++) {
                if (roomTableCodeAt[r][c] == TABLE_YES) {
                    list[i][0] = r;
                    list[i][1] = c;
                    i++;
                }
            }
        }

        // Zufälliges Vertauschen
        for (int i1 = 0; i1 < numberOfTables; i1++) {
            int i2 = (int) (Math.random() * numberOfTables);
            roomTauscheTische(list, i1, i2);
        }

        // Optimieren (1000 Mal)
        if (numberOfTables > 0) {
            for (int round = 0; round < 10000; round++) {
                int before = roomGetBewertung();
                int i1 = (int) (Math.random() * list.length);
                int i2 = (int) (Math.random() * list.length);
                roomTauscheTische(list, i1, i2);
                int after = roomGetBewertung();
                // rückgängig, falls schlechter
                if (after < before) {
                    roomTauscheTische(list, i1, i2);
                }
            }
        }

        fireModelChanged();
    }

    private void roomTauscheTische(int[][] list, int i1, int i2) {
        int r1 = list[i1][0];
        int c1 = list[i1][1];
        int r2 = list[i2][0];
        int c2 = list[i2][1];
        roomTauscheTische(r1, c1, r2, c2);
    }

    private void roomTauscheTische(int r1, int c1, int r2, int c2) {
        String vorname1 = roomVornameAt[r1][c1];
        String nachname1 = roomNachnameAt[r1][c1];
        int nummer1 = roomPersonenNummerAt[r1][c1];
        String vorname2 = roomVornameAt[r2][c2];
        String nachname2 = roomNachnameAt[r2][c2];
        int nummer2 = roomPersonenNummerAt[r2][c2];
        roomVornameAt[r1][c1] = vorname2;
        roomNachnameAt[r1][c1] = nachname2;
        roomPersonenNummerAt[r1][c1] = nummer2;
        roomVornameAt[r2][c2] = vorname1;
        roomNachnameAt[r2][c2] = nachname1;
        roomPersonenNummerAt[r2][c2] = nummer1;
    }

    public int roomGetBewertung() {
        int bewertung = 0;

        // Bewertung - Forbidden
        for (int r = 0; r < MAX_ROWS; r++) {
            for (int c = 0; c < MAX_COLS; c++) {
                // horizontal
                int n1 = roomPersonenNummerAt[r][c];
                if (r > 0) {
                    int n2 = roomPersonenNummerAt[r - 1][c];
                    if (forbiddenPair[n1][n2]) {
                        bewertung--;
                    }
                    if (allowedPair[n1][n2]) {
                        bewertung++;
                    }

                }
                // vertical
                if (c > 0) {
                    int n2 = roomPersonenNummerAt[r][c - 1];
                    if (forbiddenPair[n1][n2]) {
                        bewertung--;
                    }
                    if (allowedPair[n1][n2]) {
                        bewertung++;
                    }
                }
                // diag
                if ((r >= 1) && (c >= 1)) {
                    int n2 = roomPersonenNummerAt[r - 1][c - 1];
                    if (forbiddenPair[n1][n2])
                        bewertung--;
                    if (allowedPair[n1][n2])
                        bewertung++;
                }
                // diag
                if ((r + 1 < MAX_ROWS) && (c >= 1)) {
                    int n2 = roomPersonenNummerAt[r + 1][c - 1];
                    if (forbiddenPair[n1][n2])
                        bewertung--;
                    if (allowedPair[n1][n2])
                        bewertung++;
                }

                // diag
                if ((r >= 1) && (c + 1 < MAX_COLS)) {
                    int n2 = roomPersonenNummerAt[r - 1][c + 1];
                    if (forbiddenPair[n1][n2])
                        bewertung--;
                    if (allowedPair[n1][n2])
                        bewertung++;
                }
                // diag
                if ((r + 1 < MAX_ROWS) && (c + 1 < MAX_COLS)) {
                    int n2 = roomPersonenNummerAt[r + 1][c + 1];
                    if (forbiddenPair[n1][n2])
                        bewertung--;
                    if (allowedPair[n1][n2])
                        bewertung++;
                }
                
            }
        }

        return bewertung;
    }

    // ### IO ###
    public void write_version_7(ObjectOutputStream oos) throws Exception {

        // Version
        oos.writeInt(7);

        // Generell
        oos.writeBoolean(viewBehind);
        oos.writeBoolean(showBorder);
        oos.writeBoolean(showTableThicker);
        oos.writeInt(printState);

        // Namen [Version 7]
        oos.writeInt(MAX_PERSONS);
        for (int i = 0; i < MAX_PERSONS; i++) {
            oos.writeUTF(namesNachnameAt[i]);
            oos.writeUTF(namesVornameAt[i]);
        }

        // Verboten [Version 7]
        oos.writeUTF(forbiddenText);

        // Erlaubt [Version 7]
        oos.writeUTF(allowedText);

        // Raum [Version 7]
        for (int r = 0; r < MAX_ROWS; r++) {
            for (int c = 0; c < MAX_COLS; c++) {
                oos.writeInt(roomTableCodeAt[r][c]);
                oos.writeUTF(roomVornameAt[r][c]);
                oos.writeUTF(roomNachnameAt[r][c]);
                oos.writeInt(roomPersonenNummerAt[r][c]);
            }
        }

    }

    public void readVersion7(ObjectInputStream ois) throws Exception {

        // Generell
        viewBehind = ois.readBoolean();
        showBorder = ois.readBoolean();
        showTableThicker = ois.readBoolean();
        printState = ois.readInt();

        // Namen
        namesSetClear(ois.readInt());
        for (int rowIndex = 0; rowIndex < MAX_PERSONS; rowIndex++) {
            namesSetNachnameAtRow(ois.readUTF(), rowIndex);
            namesSetVornameAtRow(ois.readUTF(), rowIndex);
        }

        // Verboten
        forbiddenSetText(ois.readUTF());

        // Erlaubt
        allowedSetText(ois.readUTF());

        // Raum
        roomSetClear();
        for (int r = 0; r < MAX_ROWS; r++) {
            for (int c = 0; c < MAX_COLS; c++) {
                roomTableCodeAt[r][c] = ois.readInt();
                roomVornameAt[r][c] = ois.readUTF();
                roomNachnameAt[r][c] = ois.readUTF();
                roomPersonenNummerAt[r][c] = ois.readInt();
            }
        }

    }

    public void readVersion6(ObjectInputStream ois) throws Exception {

        // Generell
        viewBehind = ois.readBoolean();
        showBorder = ois.readBoolean();
        showTableThicker = ois.readBoolean();
        printState = ois.readInt();

        // Namen [Version 6]
        namesSetClear(ois.readInt());
        for (int rowIndex = 0; rowIndex < MAX_PERSONS; rowIndex++) {
            namesSetNachnameAtRow(ois.readUTF(), rowIndex);
            namesSetVornameAtRow(ois.readUTF(), rowIndex);
        }

        // Verboten [Version 6]
        forbiddenSetText(ois.readUTF());

        // Erlaubt [Version 6]
        allowedSetText(ois.readUTF());

        // Raum [Version 6]
        int rows = ois.readInt();
        int cols = ois.readInt();
        roomSetClear();

        for (int r = 0; r < rows; r++) {
            for (int c = 0; c < cols; c++) {
                roomTableCodeAt[r][c] = ois.readInt();
                roomVornameAt[r][c] = ois.readUTF();
                roomNachnameAt[r][c] = ois.readUTF();
                roomPersonenNummerAt[r][c] = ois.readInt();
            }
        }
    }

    private void readVersion5(ObjectInputStream ois) throws Exception {
        System.out.println("READING VERSION 5");

        // Namen [Version 5]
        int persons = ois.readInt();
        for (int rowIndex = 0; rowIndex < persons; rowIndex++) {
            namesSetNachnameAtRow(ois.readUTF(), rowIndex);
            namesSetVornameAtRow(ois.readUTF(), rowIndex);
        }

        // Verbotene Gruppen
        String textG = ois.readUTF();
        forbiddenSetText(textG);

        // Erwünschte Gruppen
        String textE = ois.readUTF();
        allowedSetText(textE);

        // Tables [Version 3,4,5]
        int cols = ois.readInt();
        int rows = ois.readInt();
        roomSetClear();

        for (int c = 0; c < cols; c++) {
            for (int r = 0; r < rows; r++) {
                roomTableCodeAt[r][c] = ois.readInt();

                String[] split = ois.readUTF().split(",");
                if (split.length >= 1) {
                    roomVornameAt[r][c] = split[0];
                }
                if (split.length >= 2) {
                    roomNachnameAt[r][c] = split[1];
                }
            }
        }

    }

    private void readVersion4(ObjectInputStream ois) throws Exception {
        System.out.println("READING VERSION 4");

        // Namen [Version 3,4]
        String textN = ois.readUTF();
        String[] temp = textN.split(Static.ASCII_CODE_10);
        for (int rowIndex = 0; rowIndex < temp.length; rowIndex++) {
            String[] split = temp[rowIndex].split(",");
            if (split.length >= 1) {
                namesSetVornameAtRow(split[0], rowIndex);
            }
            if (split.length >= 2) {
                namesSetNachnameAtRow(split[1], rowIndex);
            }
        }

        // Verbotene Gruppen
        String textG = ois.readUTF();
        forbiddenSetText(textG);

        // Erwünschte Gruppen
        String textE = ois.readUTF();
        allowedSetText(textE);

        // Tables [Version 3,4,5]
        int cols = ois.readInt();
        int rows = ois.readInt();
        roomSetClear();

        for (int c = 0; c < cols; c++) {
            for (int r = 0; r < rows; r++) {
                roomTableCodeAt[r][c] = ois.readInt();

                String[] split = ois.readUTF().split(",");
                if (split.length >= 1) {
                    roomVornameAt[r][c] = split[0];
                }
                if (split.length >= 2) {
                    roomNachnameAt[r][c] = split[1];
                }
            }
        }

    }

    private void readVersion3(ObjectInputStream ois) throws Exception {
        System.out.println("READING VERSION 3");

        // Namen [Version 3,4]
        String textN = ois.readUTF();
        String[] temp = textN.split(Static.ASCII_CODE_10);
        for (int rowIndex = 0; rowIndex < temp.length; rowIndex++) {
            String[] split = temp[rowIndex].split(",");
            if (split.length >= 1) {
                namesSetVornameAtRow(split[0], rowIndex);
            }
            if (split.length >= 2) {
                namesSetNachnameAtRow(split[1], rowIndex);
            }
        }

        // Verbotene Gruppen
        String textG = ois.readUTF();
        forbiddenSetText(textG);

        // Tables [Version 3,4,5]
        int cols = ois.readInt();
        int rows = ois.readInt();
        roomSetClear();

        for (int c = 0; c < cols; c++) {
            for (int r = 0; r < rows; r++) {
                roomTableCodeAt[r][c] = ois.readInt();

                String[] split = ois.readUTF().split(",");
                if (split.length >= 1) {
                    roomVornameAt[r][c] = split[0];
                }
                if (split.length >= 2) {
                    roomNachnameAt[r][c] = split[1];
                }
            }
        }
    }

    private void readVersion2(ObjectInputStream ois) throws Exception {
        System.out.println("READING VERSION 2");

        // Namen [Version 1,2]
        int persons = ois.readInt();
        for (int rowIndex = 0; rowIndex < persons; rowIndex++) {
            namesSetVornameAtRow(ois.readUTF(), rowIndex);
        }

        // Tables [Version 2]
        int cols = ois.readInt();
        int rows = ois.readInt();
        roomSetClear();

        for (int c = 0; c < cols; c++) {
            for (int r = 0; r < rows; r++) {
                roomTableCodeAt[r][c] = ois.readBoolean() ? TABLE_YES : TABLE_NO;
                roomVornameAt[r][c] = ois.readUTF();
            }
        }

    }

    private void readVersion1(ObjectInputStream ois) throws Exception {
        System.out.println("READING VERSION 1");

        // Names [Version 1,2]
        int persons = ois.readInt();
        for (int rowIndex = 0; rowIndex < persons; rowIndex++) {
            namesSetVornameAtRow(ois.readUTF(), rowIndex);
        }

        // Tables [Version 1]
        int cols = ois.readInt();
        int rows = cols;
        roomSetClear();

        for (int c = 0; c < cols; c++) {
            for (int r = 0; r < rows; r++) {
                roomTableCodeAt[r][c] = ois.readBoolean() ? TABLE_YES : TABLE_NO;
                roomVornameAt[r][c] = ois.readUTF();
            }
        }

    }

    public void readFromStream(ObjectInputStream ois) throws Exception {
        int version = ois.readInt();
        System.out.println("READ GLOBAL VERSION " + version);
        if (version == 1) {
            readVersion1(ois);
            return;
        }
        if (version == 2) {
            readVersion2(ois);
            return;
        }
        if (version == 3) {
            readVersion3(ois);
            return;
        }
        if (version == 4) {
            readVersion4(ois);
            return;
        }
        if (version == 5) {
            readVersion5(ois);
            return;
        }
        if (version == 6) {
            readVersion6(ois);
            return;
        }
        if (version == 7) {
            readVersion7(ois);
            return;
        }

        throw new Exception("Dateiversion " + version + " ist unbekannt!");
    }

    // ### OTHERS ###
    public void addModelChangeListener(ModelChangeListener l) {
        listenersAll.add(l);
        fireModelChanged();
    }

    public void fireModelChanged() {
        for (ModelChangeListener l : listenersAll) {
            l.modelChanged();
        }
    }

    // ---- NAMEN ----
    public void namesSetClear(int size) {
        MAX_PERSONS = size;
        namesVornameAt = new String[MAX_PERSONS];
        namesNachnameAt = new String[MAX_PERSONS];
        for (int i = 0; i < MAX_PERSONS; i++) {
            namesVornameAt[i] = "";
            namesNachnameAt[i] = "";
        }
    }

    public int namesGetNumberOfValidPersons() {
        int count = 0;

        for (int rowIndex = 0; rowIndex < MAX_PERSONS; rowIndex++) {
            String sNachname = namesNachnameAt[rowIndex];
            String sVorname = namesVornameAt[rowIndex];
            if (sNachname.equals("") && sVorname.equals("")) {
                continue;
            }
            count++;
        }

        return count;
    }

    public void namesSetVornameAtRow(String vorname, int rowIndex) {
        vorname = namesNormalize(vorname);
        namesVornameAt[rowIndex] = vorname;
        fireRowChanged(rowIndex);
        fireModelChanged();
    }

    public void namesSetNachnameAtRow(String nachname, int rowIndex) {
        nachname = namesNormalize(nachname);
        namesNachnameAt[rowIndex] = nachname;
        fireRowChanged(rowIndex);
        fireModelChanged();
    }

    // ### TABLE MODEL OF NAMES ###
    @Override
    public int getRowCount() {
        return namesVornameAt.length;
    }

    @Override
    public int getColumnCount() {
        return 3;
    }

    @Override
    public String getColumnName(int columnIndex) {
        if (columnIndex == 0) {
            return HEADER_NUMMER;
        }
        if (columnIndex == 1) {
            return HEADER_NACHNAME;
        }
        if (columnIndex == 2) {
            return HEADER_VORNAME;
        }
        throw new UnsupportedOperationException("INVALID columnIndex " + columnIndex);
    }

    @Override
    public Class<?> getColumnClass(int columnIndex) {
        if (columnIndex == 0) {
            return Integer.class;
        }
        if (columnIndex == 1) {
            return String.class;
        }
        if (columnIndex == 2) {
            return String.class;
        }
        throw new UnsupportedOperationException("INVALID columnIndex " + columnIndex);
    }

    @Override
    public boolean isCellEditable(int rowIndex, int columnIndex) {
        return columnIndex > 0;
    }

    @Override
    public Object getValueAt(int rowIndex, int columnIndex) {
        if (columnIndex == 0) {
            return rowIndex + 1;
        }
        if (columnIndex == 1) {
            return namesNachnameAt[rowIndex];
        }
        if (columnIndex == 2) {
            return namesVornameAt[rowIndex];
        }
        throw new UnsupportedOperationException("INVALID columnIndex " + columnIndex);
    }

    @Override
    public void setValueAt(Object aValue, int rowIndex, int columnIndex) {
        if (columnIndex <= 0) {
            return;
        }
        if (columnIndex >= 3) {
            return;
        }
        if (columnIndex == 1) {
            namesSetNachnameAtRow((String) aValue, rowIndex);
        }
        if (columnIndex == 2) {
            namesSetVornameAtRow((String) aValue, rowIndex);
        }

        fireRowChanged(rowIndex);
        fireModelChanged();
    }

    private void fireRowChanged(int rowIndex) {
        TableModelEvent tme = new TableModelEvent(this, rowIndex);
        for (TableModelListener tml : listenersNames) {
            tml.tableChanged(tme);
        }
    }

    @Override
    public void addTableModelListener(TableModelListener l) {
        listenersNames.add(l);
    }

    @Override
    public void removeTableModelListener(TableModelListener l) {
        listenersNames.remove(l);
    }

    public boolean getForbiddenNeihbours(int r1, int c1, int r2, int c2) {
        if ((r1 < 0) || (r1 >= MAX_ROWS)) {
            return false;
        }
        if ((r2 < 0) || (r2 >= MAX_ROWS)) {
            return false;
        }
        if ((c1 < 0) || (c1 >= MAX_COLS)) {
            return false;
        }
        if ((c2 < 0) || (c2 >= MAX_COLS)) {
            return false;
        }
        int n1 = roomPersonenNummerAt[r1][c1];
        int n2 = roomPersonenNummerAt[r2][c2];
        return forbiddenPair[n1][n2];
    }

    public boolean getAllowedNeihbours(int r1, int c1, int r2, int c2) {
        if ((r1 < 0) || (r1 >= MAX_ROWS)) {
            return false;
        }
        if ((r2 < 0) || (r2 >= MAX_ROWS)) {
            return false;
        }
        if ((c1 < 0) || (c1 >= MAX_COLS)) {
            return false;
        }
        if ((c2 < 0) || (c2 >= MAX_COLS)) {
            return false;
        }
        int n1 = roomPersonenNummerAt[r1][c1];
        int n2 = roomPersonenNummerAt[r2][c2];
        return allowedPair[n1][n2];
    }

    public void readFromFile(File file) throws Exception {
        FileInputStream fis = new FileInputStream(file);
        BufferedInputStream bis = new BufferedInputStream(fis);
        ObjectInputStream ois = new ObjectInputStream(bis);
        readFromStream(ois);
        ois.close();
    }

    private String namesNormalize(String name) {
        name = name.trim();
        return name;
    }

    public void setFileToRemember(File file) throws Exception {
        // Save Path
        lastOpenedDirectory = file.getParentFile();
    }

    protected File getDirectoryLastOpened() {
        return lastOpenedDirectory;
    }

}
