package sitzplangenerator;

import java.awt.*;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.awt.print.PrinterException;
import javax.swing.JPanel;

/**
 *
 * @author B. A. Bartsch
 */
public class ViewPanelDraw extends JPanel implements MouseMotionListener, MouseListener, Printable, ModelChangeListener {

    private final Model model;
    private final Font fontText = new Font("SansSerif", Font.PLAIN, 10);
    //private final Font fontText2 = new Font("SansSerif", Font.PLAIN, 8);
    private final Font fontHeader = new Font("SansSerif", Font.BOLD, 20);

    @SuppressWarnings({"LeakingThisInConstructor", "OverridableMethodCallInConstructor"})
    public ViewPanelDraw(Model m) {
        model = m;
        addMouseMotionListener(this);
        addMouseListener(this);
        model.addModelChangeListener(this);
    }

    @Override
    public void modelChanged() {
        repaint();
    }

    @Override
    public void paint(Graphics g) {
        super.paint(g);
        paintRoom(g, false);
    }

    private void paintRoom(Graphics g, boolean isPrintJob) {
        int ROOM_X = model.roomGetDeltaX();
        int ROOM_Y = model.roomGetDeltaY();
        int TABLE_W = model.roomGetTableWidth();
        int TABLE_H = model.roomGetTableHeight();
        int MAX_COLS = model.roomGetColumns();
        int MAX_ROWS = model.roomGetRows();
        int ROOM_W = MAX_COLS * TABLE_W;
        int ROOM_H = MAX_ROWS * TABLE_H;
        //int COLSM1 = MAX_COLS - 1;
        //int ROWSM1 = MAX_ROWS - 1;

        // Zeichnen der Tafel
        g.setColor(Color.BLACK);
        g.setFont(fontHeader);
        relativeStringInRectangle(g, model.getHeader(), ROOM_X, ROOM_Y, ROOM_W, TABLE_H);

        // Zeichnen des Randes
        if (model.getShowBorder()) {
            g.setColor(Color.GRAY);
            relativeRectangle(g, ROOM_X, ROOM_Y, ROOM_W, ROOM_H);
            relativeRectangle(g, ROOM_X - 1, ROOM_Y - 1, ROOM_W + 2, ROOM_H + 2);
        }

        // Zeichnen der Tische
        g.setColor(Color.BLACK);
        for (int x = 0; x < MAX_COLS; x++) {
            for (int y = 0; y < MAX_ROWS; y++) {
                int tableCode = model.roomGetTableCodeAt(y, x);
                relativeTable(g, isPrintJob, tableCode, ROOM_X + TABLE_W * x, ROOM_Y + TABLE_H * y, TABLE_W, TABLE_H);
            }
        }

        if (!isPrintJob) {
            int stroke = 2;
            // Zeichnen der Präferenzen
            g.setColor(Color.RED);
            for (int r = 0; r < MAX_ROWS; r++) {
                for (int c = 0; c < MAX_COLS; c++) {
                    // diagonal
                    int x1 = ROOM_X + TABLE_W / 2 + TABLE_W * c;
                    int y1 = ROOM_Y + TABLE_H / 2 + TABLE_H * r;
                    if (model.getForbiddenNeihbours(r, c, r - 1, c - 1)) {
                        int x2 = ROOM_X + TABLE_W / 2 + TABLE_W * (c - 1);
                        int y2 = ROOM_Y + TABLE_H / 2 + TABLE_H * (r - 1);
                        relativeLine(g, x1, y1, x2, y2, stroke);
                    }
                    if (model.getForbiddenNeihbours(r, c, r + 1, c - 1)) {
                        int x2 = ROOM_X + TABLE_W / 2 + TABLE_W * (c - 1);
                        int y2 = ROOM_Y + TABLE_H / 2 + TABLE_H * (r + 1);
                        relativeLine(g, x1, y1, x2, y2, stroke);
                    }
                    if (model.getForbiddenNeihbours(r, c, r - 1, c + 1)) {
                        int x2 = ROOM_X + TABLE_W / 2 + TABLE_W * (c + 1);
                        int y2 = ROOM_Y + TABLE_H / 2 + TABLE_H * (r - 1);
                        relativeLine(g, x1, y1, x2, y2, stroke);
                    }
                    if (model.getForbiddenNeihbours(r, c, r + 1, c + 1)) {
                        int x2 = ROOM_X + TABLE_W / 2 + TABLE_W * (c + 1);
                        int y2 = ROOM_Y + TABLE_H / 2 + TABLE_H * (r + 1);
                        relativeLine(g, x1, y1, x2, y2, stroke);
                    }
                    // horizontal & vertikal
                    if (model.getForbiddenNeihbours(r, c, r - 1, c)) {
                        int x2 = ROOM_X + TABLE_W / 2 + TABLE_W * (c);
                        int y2 = ROOM_Y + TABLE_H / 2 + TABLE_H * (r - 1);
                        relativeLine(g, x1, y1, x2, y2, stroke);
                    }
                    if (model.getForbiddenNeihbours(r, c, r, c - 1)) {
                        int x2 = ROOM_X + TABLE_W / 2 + TABLE_W * (c - 1);
                        int y2 = ROOM_Y + TABLE_H / 2 + TABLE_H * (r);
                        relativeLine(g, x1, y1, x2, y2, stroke);
                    }
                }
            }
            g.setColor(Color.GREEN);
            for (int r = 0; r < MAX_ROWS; r++) {
                for (int c = 0; c < MAX_COLS; c++) {
                    int x1 = ROOM_X + TABLE_W / 2 + TABLE_W * c;
                    int y1 = ROOM_Y + TABLE_H / 2 + TABLE_H * r;
                    if (model.getAllowedNeihbours(r, c, r - 1, c - 1)) {
                        int x2 = ROOM_X + TABLE_W / 2 + TABLE_W * (c - 1);
                        int y2 = ROOM_Y + TABLE_H / 2 + TABLE_H * (r - 1);
                        relativeLine(g, x1, y1, x2, y2, stroke);
                    }
                    if (model.getAllowedNeihbours(r, c, r + 1, c - 1)) {
                        int x2 = ROOM_X + TABLE_W / 2 + TABLE_W * (c - 1);
                        int y2 = ROOM_Y + TABLE_H / 2 + TABLE_H * (r + 1);
                        relativeLine(g, x1, y1, x2, y2, stroke);
                    }
                    if (model.getAllowedNeihbours(r, c, r - 1, c + 1)) {
                        int x2 = ROOM_X + TABLE_W / 2 + TABLE_W * (c + 1);
                        int y2 = ROOM_Y + TABLE_H / 2 + TABLE_H * (r - 1);
                        relativeLine(g, x1, y1, x2, y2, stroke);
                    }
                    if (model.getAllowedNeihbours(r, c, r + 1, c + 1)) {
                        int x2 = ROOM_X + TABLE_W / 2 + TABLE_W * (c + 1);
                        int y2 = ROOM_Y + TABLE_H / 2 + TABLE_H * (r + 1);
                        relativeLine(g, x1, y1, x2, y2, stroke);
                    }
                    // horizontal & vertikal
                    if (model.getAllowedNeihbours(r, c, r - 1, c)) {
                        int x2 = ROOM_X + TABLE_W / 2 + TABLE_W * (c);
                        int y2 = ROOM_Y + TABLE_H / 2 + TABLE_H * (r - 1);
                        relativeLine(g, x1, y1, x2, y2, stroke);
                    }
                    if (model.getAllowedNeihbours(r, c, r, c - 1)) {
                        int x2 = ROOM_X + TABLE_W / 2 + TABLE_W * (c - 1);
                        int y2 = ROOM_Y + TABLE_H / 2 + TABLE_H * (r);
                        relativeLine(g, x1, y1, x2, y2, stroke);
                    }
                    /*if (model.getAllowedNeihbours(r, c, r - 1, c)) {
                        relativeFillOval(g, ROOM_X + TABLE_W * c + TABLE_W / 2 - 5, ROOM_Y + TABLE_H * r - 5, 10, 10);
                    }
                    if (model.getAllowedNeihbours(r, c, r, c - 1)) {
                        relativeFillOval(g, ROOM_X + TABLE_W * c - 5, ROOM_Y + TABLE_H * r + TABLE_H / 2 - 5, 10, 10);
                    }*/
                }
            }

            // Zeichnen der Kollisionen
            g.setFont(fontText);
            g.setColor(Color.BLACK);
            g.drawString("Sitzplanbewertung = " + model.roomGetBewertung(), ROOM_X, ROOM_Y - 5);

            // Zeichnen Verschiebungslinie
            if (model.roomGetMouseStartRightClickX() >= 0) {
                g.setColor(Color.BLUE);
                int x1 = model.roomGetMouseStartRightClickX();
                int y1 = model.roomGetMouseStartRightClickY();
                int x2 = model.roomGetMouseCurrentX();
                int y2 = model.roomGetMouseCurrentY();
                g.drawLine(x1, y1, x2, y2);
            }
        }

        // Zeichnen der Namen
        g.setFont(fontText);
        g.setColor(Color.BLACK);
        for (int row = 0; row < MAX_ROWS; row++) {
            for (int col = 0; col < MAX_COLS; col++) {
                String vorname = model.roomGetTableVornameAt(row, col);
                String nachname = model.roomGetTableNachnameAt(row, col);
                int personenID = model.roomGetTablePersonenNummerAt(row, col);
                g.setFont(Static.FONT_NAME);
                relativeNameInRectangle(g, vorname, nachname, ROOM_X + TABLE_W * col, ROOM_Y + TABLE_H * row, TABLE_W, TABLE_H);
                if (!vorname.isEmpty() || !nachname.isEmpty()) {
                    if (!isPrintJob) {
                        g.setColor(Color.BLUE);
                        g.setFont(Static.FONT_NUMBER);
                        relativeNumberInRectangle(g, personenID, ROOM_X + TABLE_W * col, ROOM_Y + TABLE_H * row, TABLE_W, TABLE_H);
                        g.setColor(Color.BLACK);
                    }
                }
            }
        }

    }

    private int relativeInvertX(int x) {
        if (model.getViewSchueler()) {
            return x;
        }
        x = x - model.roomGetDeltaX();
        x = model.roomGetColumns() * model.roomGetTableWidth() - x;
        x = x + model.roomGetDeltaX();
        return x;
    }

    private int relativeInvertY(int y) {
        if (model.getViewSchueler()) {
            return y;
        }
        y = y - model.roomGetDeltaY();
        y = model.roomGetRows() * model.roomGetTableHeight() - y;
        y = y + model.roomGetDeltaY();
        return y;
    }

    private void relativeNumberInRectangle(Graphics g, int personenID, int x, int y, int w, int h) {
        int x1 = x;
        int y1 = y;
        int x2 = x + w;
        int y2 = y + h;
        x1 = relativeInvertX(x1);
        y1 = relativeInvertY(y1);
        x2 = relativeInvertX(x2);
        y2 = relativeInvertY(y2);
        x = Math.min(x1, x2);
        y = Math.min(y1, y2);
        drawUpperLeftString(g, "" + personenID, x, y, w, h);
    }

    private void relativeNameInRectangle(Graphics g, String vorname, String nachname, int x, int y, int w, int h) {
        int x1 = x;
        int y1 = y;
        int x2 = x + w;
        int y2 = y + h;
        x1 = relativeInvertX(x1);
        y1 = relativeInvertY(y1);
        x2 = relativeInvertX(x2);
        y2 = relativeInvertY(y2);
        x = Math.min(x1, x2);
        y = Math.min(y1, y2);

        if (model.getPrintState() == Model.PRINT_VORNAME) {
            drawCenteredString(g, vorname, x, y, w, h);
            return;
        }
        if (model.getPrintState() == Model.PRINT_NACHNAME) {
            drawCenteredString(g, nachname, x, y, w, h);
            return;
        }
        if (model.getPrintState() == Model.PRINT_VORNAME_NACHNAME) {
            drawCenteredString(g, vorname, x, y + (h * 1) / 3, w, h / 3);
            drawCenteredString(g, nachname, x, y + (h * 2) / 3, w, h / 3);

            //drawCenteredString(g, vorname, x, y, w, h / 2);
            //drawCenteredString(g, nachname, x, y + h / 2, w, h / 2);
            return;
        }
        if (model.getPrintState() == Model.PRINT_NACHNAME_VORNAME) {
            drawCenteredString(g, nachname, x, y + (h * 1) / 3, w, h / 3);
            drawCenteredString(g, vorname, x, y + (h * 2) / 3, w, h / 3);
            //drawCenteredString(g, nachname, x, y, w, h / 2);
            //drawCenteredString(g, vorname, x, y + h / 2, w, h / 2);
            return;
        }

        //throw new UnsupportedOperationException("Not yet implemented");
    }

    private void relativeStringInRectangle(Graphics g, String s, int x1, int y1, int w, int h) {
        int x2 = x1 + w;
        int y2 = y1 + h;
        x1 = relativeInvertX(x1);
        y1 = relativeInvertY(y1);
        x2 = relativeInvertX(x2);
        y2 = relativeInvertY(y2);
        int xMin = Math.min(x1, x2);
        int yMin = Math.min(y1, y2);
        //int xMax = Math.max(x1, x2);
        //int yMax = Math.max(y1, y2);
        drawCenteredString(g, s, xMin, yMin, w, h);
    }

    private void relativeRectangle(Graphics g, int x1, int y1, int w, int h) {
        int x2 = x1 + w;
        int y2 = y1 + h;
        x1 = relativeInvertX(x1);
        y1 = relativeInvertY(y1);
        x2 = relativeInvertX(x2);
        y2 = relativeInvertY(y2);
        int xMin = Math.min(x1, x2);
        int yMin = Math.min(y1, y2);
        g.drawRect(xMin, yMin, w, h);
    }

    private void relativeFillRect(Graphics g, int x1, int y1, int w, int h) {
        int x2 = x1 + w;
        int y2 = y1 + h;
        x1 = relativeInvertX(x1);
        y1 = relativeInvertY(y1);
        x2 = relativeInvertX(x2);
        y2 = relativeInvertY(y2);
        int xMin = Math.min(x1, x2);
        int yMin = Math.min(y1, y2);
        g.fillRect(xMin, yMin, w, h);
    }

    private void relativeFillOval(Graphics g, int x1, int y1, int w, int h) {
        int x2 = x1 + w;
        int y2 = y1 + h;
        x1 = relativeInvertX(x1);
        y1 = relativeInvertY(y1);
        x2 = relativeInvertX(x2);
        y2 = relativeInvertY(y2);
        int xMin = Math.min(x1, x2);
        int yMin = Math.min(y1, y2);
        g.fillOval(xMin, yMin, w, h);
    }

    private void relativeLine(Graphics g, int x1, int y1, int x2, int y2, int stroke) {
        //int x2 = x1 + w;
        //int y2 = y1 + h;
        x1 = relativeInvertX(x1);
        y1 = relativeInvertY(y1);
        x2 = relativeInvertX(x2);
        y2 = relativeInvertY(y2);
        ((Graphics2D) g).setStroke(new BasicStroke(stroke));
        g.drawLine(x1, y1, x2, y2);
    }

    private void relativeTable(Graphics g, boolean print, int tableCode, int x, int y, int w, int h) {
        if (tableCode == Model.TABLE_NO) {
            return;
        }

        // rechteck
        relativeRectangle(g, x, y, w, h);
        if (model.roomGetTableThickerActivated()) {
            relativeRectangle(g, x - 1, y - 1, w, h);
        }

        // kreuz
        if ((!print) && (tableCode == Model.TABLE_BLOCKED)) {
            relativeLine(g, x, y, x + w, y + h, 1);
            relativeLine(g, x, y + h, x + w, y, 1);
            //g.drawLine(x,y,x+w, y+h);
            //g.drawLine(x,y+h,x+w, y);
        }
    }

    public void drawCenteredString(Graphics g, String s, int x, int y, int w, int h) {
        /*g.setFont(fontText);
         if (s.length() > 8) {
         g.setFont(fontText2);
         }*/
        FontMetrics fm = g.getFontMetrics();
        int dx = (w - fm.stringWidth(s)) / 2;
        int dy = (fm.getAscent() + (h - (fm.getAscent() + fm.getDescent())) / 2);
        g.drawString(s, x + dx, y + dy);
    }

    public void drawUpperLeftString(Graphics g, String s, int x, int y, int w, int h) {
        FontMetrics fm = g.getFontMetrics();
//        int dx = (w - fm.stringWidth(s)) / 2;
        int dx = 1;
//        int dy = (fm.getAscent() + (h - (fm.getAscent() + fm.getDescent())) / 2);
        int dy = fm.getAscent();
        g.drawString(s, x + dx, y + dy);
    }

    @Override
    public void mouseDragged(MouseEvent e) {
        model.roomSetMouseCurrent(e.getX(), e.getY());
    }

    @Override
    public void mouseMoved(MouseEvent e) {
        repaint();
    }

    @Override
    public void mouseClicked(MouseEvent e) {
    }

    @Override
    public void mousePressed(MouseEvent e) {
        Point p = e.getPoint();

        if (e.getButton() == MouseEvent.BUTTON1) {
            model.roomSetMouseStartLeftClick(e.getX(), e.getY());
        }

        if (e.getButton() == MouseEvent.BUTTON3) {
            model.roomSetMouseStartRightClick(e.getX(), e.getY());
        }
    }

    @Override
    public void mouseReleased(MouseEvent e) {
        if (e.getButton() == MouseEvent.BUTTON3) {
            model.roomSetMouseReleasedRight(e.getX(), e.getY());
        }
    }

    @Override
    public void mouseEntered(MouseEvent e) {
    }

    @Override
    public void mouseExited(MouseEvent e) {
    }

    @Override
    public int print(Graphics graphics, PageFormat pageFormat, int pageIndex) throws PrinterException {
        if (pageIndex >= 1) {
            return Printable.NO_SUCH_PAGE;
        }
        paintRoom(graphics, true);
        return Printable.PAGE_EXISTS;
    }

}
