package sitzplangenerator;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.print.PageFormat;
import java.awt.print.Paper;
import java.awt.print.PrinterJob;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.ObjectOutputStream;
import java.io.UnsupportedEncodingException;
import java.util.LinkedList;
import javax.print.attribute.HashPrintRequestAttributeSet;
import javax.print.attribute.standard.Chromaticity;
import javax.print.attribute.standard.MediaPrintableArea;
import javax.print.attribute.standard.MediaSizeName;
import javax.print.attribute.standard.OrientationRequested;
import javax.print.attribute.standard.PrintQuality;
import javax.swing.*;
import javax.swing.border.TitledBorder;

/**
 *
 * @author B. A. Bartsch
 */
public class ViewPanelMain extends JPanel implements ActionListener, KeyListener, ModelChangeListener {

    private final Model model = new Model(50); // Platz für 50 Personen
    private final ViewPanelDraw draw = new ViewPanelDraw(model);

    private final JButton buttonOpen = new JButton("");
    private final JButton buttonSave = new JButton("");
    private final JButton buttonPrint = new JButton("");
    private final JButton buttonGenerate = new JButton("");
    private final JButton buttonImportTxt = new JButton("");
    private final JCheckBox showBorder = new JCheckBox("Rand?");
    private final JCheckBox showTischThick = new JCheckBox("Fett?");
    private final JFileChooser fileChooser = new JFileChooser();
    private final JFileChooser fileChooserCSV = new JFileChooser();

    private final JRadioButton radioView1 = new JRadioButton("Sicht 1 (Schüler)");
    private final JRadioButton radioView2 = new JRadioButton("Sicht 2 (Lehrer)");
    private final ButtonGroup groupView12 = new ButtonGroup();

    private final JRadioButton radioPrint1 = new JRadioButton("V");
    private final JRadioButton radioPrint2 = new JRadioButton("N");
    private final JRadioButton radioPrint3 = new JRadioButton("V/N");
    private final JRadioButton radioPrint4 = new JRadioButton("N/V");
    private final ButtonGroup groupPrint1234 = new ButtonGroup();

    // Namen
    private final TitledBorder namesBorder = BorderFactory.createTitledBorder("Namensliste");
    private final JTable namesTable = new JTable(model);
    private final JScrollPane namesScroll = new JScrollPane(namesTable);

    private final String titlePartnerV = "Verbotene Gruppen (z.B: 1,4,7) = ";
    private final TitledBorder forbiddenBorder = BorderFactory.createTitledBorder(titlePartnerV + "0");
    private final JTextArea forbiddenTextArea = new JTextArea();
    private final JScrollPane forbiddenScroll = new JScrollPane(forbiddenTextArea);

    private final String titlePartnerE = "Erwünschte Partner (z.B: 2,5) = ";
    private final TitledBorder allowedBorder = BorderFactory.createTitledBorder(titlePartnerE + "0");
    private final JTextArea allowedTextArea = new JTextArea();
    private final JScrollPane allowedScroll = new JScrollPane(allowedTextArea);

    // Aktuelle Datei
    private final JLabel header = new JLabel("Datei: ");

    @SuppressWarnings({"LeakingThisInConstructor", "OverridableMethodCallInConstructor"})
    public ViewPanelMain() {
        super(new BorderLayout());

        namesScroll.setBorder(namesBorder);
        namesTable.setColumnSelectionAllowed(false);
        namesTable.setRowSelectionAllowed(false);
        namesScroll.setPreferredSize(new Dimension(300, 100));

        buttonOpen.setIcon(Static.IMAGE_OPEN);
        buttonSave.setIcon(Static.IMAGE_SAVE);
        buttonPrint.setIcon(Static.IMAGE_PRINT);
        buttonGenerate.setIcon(Static.IMAGE_SHUFFLE);
        buttonImportTxt.setIcon(Static.IMAGE_IMPORT);

        buttonOpen.setToolTipText("Öffnen...");
        buttonSave.setToolTipText("Speichern...");
        buttonPrint.setToolTipText("Drucken...");
        buttonGenerate.setToolTipText("Sitzplan generieren...");
        buttonImportTxt.setToolTipText("Text-Datei importieren. Erste Zeile: Nachname;Vorname");

        fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
        fileChooser.setFileHidingEnabled(true);
        fileChooser.setMultiSelectionEnabled(false);
        fileChooser.setFileFilter(new FileFilterSPG());

        fileChooserCSV.setFileSelectionMode(JFileChooser.FILES_ONLY);
        fileChooserCSV.setFileHidingEnabled(true);
        fileChooserCSV.setMultiSelectionEnabled(false);
        fileChooserCSV.setFileFilter(new FileFilterTXT());

        groupView12.add(radioView1);
        groupView12.add(radioView2);

        groupPrint1234.add(radioPrint1);
        groupPrint1234.add(radioPrint2);
        groupPrint1234.add(radioPrint3);
        groupPrint1234.add(radioPrint4);

        forbiddenTextArea.setLineWrap(false);
        forbiddenTextArea.setRows(8);
        forbiddenScroll.setBorder(forbiddenBorder);

        allowedTextArea.setLineWrap(false);
        allowedTextArea.setRows(8);
        allowedScroll.setBorder(allowedBorder);

        JToolBar barH1 = new JToolBar(JToolBar.HORIZONTAL);
        barH1.setAlignmentX(LEFT_ALIGNMENT);
        barH1.setFloatable(false);
        barH1.add(buttonOpen);
        barH1.add(buttonSave);
        barH1.add(buttonPrint);
        barH1.add(buttonImportTxt);
        barH1.add(buttonGenerate);

        JToolBar barH2 = new JToolBar(JToolBar.HORIZONTAL);
        barH2.setAlignmentX(LEFT_ALIGNMENT);
        barH2.setFloatable(false);
        barH2.add(showBorder);
        barH2.add(showTischThick);

        JToolBar barH3 = new JToolBar(JToolBar.HORIZONTAL);
        barH3.setAlignmentX(LEFT_ALIGNMENT);
        barH3.setFloatable(false);
        barH3.add(radioView1);
        barH3.add(radioView2);

        JToolBar barH4 = new JToolBar(JToolBar.HORIZONTAL);
        barH4.setAlignmentX(LEFT_ALIGNMENT);
        barH4.setFloatable(false);
        barH4.add(radioPrint1);
        barH4.add(radioPrint2);
        barH4.add(radioPrint3);
        barH4.add(radioPrint4);

        JToolBar barH5 = new JToolBar(JToolBar.HORIZONTAL);
        barH5.setAlignmentX(LEFT_ALIGNMENT);
        barH5.setFloatable(false);
        barH5.add(namesScroll);

        JToolBar barV = new JToolBar(JToolBar.VERTICAL);
        barV.setFloatable(false);
        barV.add(barH1);
        barV.add(barH2);
        barV.add(barH3);
        barV.add(barH4);
        barV.add(barH5);

        JPanel panel5 = new JPanel(new BorderLayout());
        panel5.add(barV, BorderLayout.CENTER);
        panel5.add(forbiddenScroll, BorderLayout.SOUTH);

        JPanel panel6 = new JPanel(new BorderLayout());
        panel6.add(panel5, BorderLayout.CENTER);
        panel6.add(allowedScroll, BorderLayout.SOUTH);

        this.add(draw, BorderLayout.CENTER);
        this.add(panel6, BorderLayout.WEST);
        this.add(header, BorderLayout.NORTH);

        buttonGenerate.addActionListener(this);
        buttonImportTxt.addActionListener(this);
        buttonPrint.addActionListener(this);
        radioView1.addActionListener(this);
        radioView2.addActionListener(this);
        buttonSave.addActionListener(this);
        buttonOpen.addActionListener(this);
        showBorder.addActionListener(this);
        showTischThick.addActionListener(this);
        forbiddenTextArea.addKeyListener(this);
        allowedTextArea.addKeyListener(this);
        radioPrint1.addActionListener(this);
        radioPrint2.addActionListener(this);
        radioPrint3.addActionListener(this);
        radioPrint4.addActionListener(this);

        model.addModelChangeListener(this); // causes fire
    }

    // ### BUTTON CLICKS ###
    @Override
    @SuppressWarnings("CallToPrintStackTrace")
    public void actionPerformed(ActionEvent e) {
        try {
            Object src = e.getSource();
            if (src == buttonOpen) {
                buttonOpenChooser();
            }
            if (src == buttonSave) {
                buttonSave();
            }
            if (src == buttonGenerate) {
                buttonGenerate();
            }
            if (src == buttonImportTxt) {
                buttonImportCSV();
            }
            if (src == buttonPrint) {
                buttonPrint();
            }
            if ((src == radioView1) || (src == radioView2)) {
                model.setViewBehind(radioView1.isSelected());
            }
            if ((src == radioPrint1) && (radioPrint1.isSelected())) {
                model.setPrintState(Model.PRINT_VORNAME);
            }
            if ((src == radioPrint2) && (radioPrint2.isSelected())) {
                model.setPrintState(Model.PRINT_NACHNAME);
            }
            if ((src == radioPrint3) && (radioPrint3.isSelected())) {
                model.setPrintState(Model.PRINT_VORNAME_NACHNAME);
            }
            if ((src == radioPrint4) && (radioPrint4.isSelected())) {
                model.setPrintState(Model.PRINT_NACHNAME_VORNAME);
            }
            if (src == showBorder) {
                model.setShowBorder(showBorder.isSelected());
            }
            if (src == showTischThick) {
                model.roomSetTableThicker(showTischThick.isSelected());
            }
        } catch (Exception ex) {
            JOptionPane.showMessageDialog(this, "Programmfehler: " + ex.getLocalizedMessage());
            ex.printStackTrace();
        }
    }

    @Override
    public void keyReleased(KeyEvent e) {
        if (e.getSource() == forbiddenTextArea) {
            model.forbiddenSetText(forbiddenTextArea.getText());
        }
        if (e.getSource() == allowedTextArea) {
            model.allowedSetText(allowedTextArea.getText());
        }
    }

    @Override
    public void keyTyped(KeyEvent e) {
    }

    @Override
    public void keyPressed(KeyEvent e) {
    }

    private void buttonGenerate() throws Exception {
        model.roomSetTableRandomSeatingOrder();
    }

    private void buttonPrint() throws Exception {
        HashPrintRequestAttributeSet attr = new HashPrintRequestAttributeSet();
        attr.add(MediaSizeName.ISO_A4);
        attr.add(new MediaPrintableArea(5f, 5f, 210f - 2 * 5f, 297f - 2 * 5f, MediaPrintableArea.MM));
        attr.add(OrientationRequested.PORTRAIT);
        attr.add(PrintQuality.NORMAL);
        attr.add(Chromaticity.MONOCHROME);

        PrinterJob pjob = PrinterJob.getPrinterJob();
        PageFormat pageFormat = pjob.defaultPage();
        Paper paper = pageFormat.getPaper();
        paper.setImageableArea(0, 0, paper.getWidth(), paper.getHeight());
        pageFormat.setPaper(paper);

        if (pjob.printDialog(attr) == false) {
            return;
        }


        pjob.setPrintable(draw, pjob.defaultPage(pageFormat));

        //pjob.setPrintable(draw, pjob.defaultPage());

        pjob.print();
    }

    private void buttonSave() throws Exception {
        buttonSave.requestFocus();
        // open last used directory
        File lastOpenedDirectory = model.getDirectoryLastOpened();
        if (lastOpenedDirectory != null) {
            fileChooser.setCurrentDirectory(lastOpenedDirectory);
        }

        if (fileChooser.showSaveDialog(this) != JFileChooser.APPROVE_OPTION) {
            return;
        }

        File file = fileChooser.getSelectedFile();

        if (file == null) {
            return;
        }

        File parent = file.getParentFile();
        String child = file.getName().toLowerCase();
        if (!child.endsWith(".spg")) {
            child = child + ".spg";
        }

        file = new File(parent, child);

        if (file.exists()) {
            int option = JOptionPane.showConfirmDialog(this, "" + file + "",
                    "Möchten Sie die folgende Datei überschreiben?",
                    JOptionPane.YES_NO_OPTION);
            if (option != JOptionPane.YES_OPTION) {
                return;
            }
        }

        model.setFileToRemember(file);

        FileOutputStream fos = new FileOutputStream(file);
        BufferedOutputStream bos = new BufferedOutputStream(fos);
        ObjectOutputStream oos = new ObjectOutputStream(bos);
        model.write_version_7(oos);
        oos.close();
    }

    private void buttonOpenChooser() throws Exception {
        // open last used directory
        File lastOpenedDirectory = model.getDirectoryLastOpened();
        if (lastOpenedDirectory != null) {
            fileChooser.setCurrentDirectory(lastOpenedDirectory);
        }

        if (fileChooser.showOpenDialog(this) != JFileChooser.APPROVE_OPTION) {
            return;
        }

        File file = fileChooser.getSelectedFile();
        buttonOpenFile(file);
    }

    private void buttonOpenFile(File file) throws Exception {

        if (file == null) {
            return;
        }

        if (!file.exists()) {
            return;
        }

        model.setClearAll();
        forbiddenTextArea.setText("");
        allowedTextArea.setText("");

        model.readFromFile(file);

        header.setText("Datei: " + file.getAbsolutePath());
        forbiddenTextArea.setText(model.getForbiddenText());
        allowedTextArea.setText(model.getAllowedText());
        model.fireModelChanged();
    }

    @Override
    public void modelChanged() {
        int forbiddenPairs = model.getForbiddenCount();
        forbiddenBorder.setTitle(titlePartnerV + forbiddenPairs);
        forbiddenScroll.repaint();

        int allowedPairs = model.getAllowedCount();
        allowedBorder.setTitle(titlePartnerE + allowedPairs);
        allowedScroll.repaint();

        int numberOfPersons = model.namesGetNumberOfValidPersons();
        namesBorder.setTitle("Namensliste (" + numberOfPersons + ")");
        namesScroll.repaint();

        showBorder.setSelected(model.getShowBorder());
        showTischThick.setSelected(model.getShowTableThicker());
        radioView1.setSelected(model.getViewSchueler());
        radioView2.setSelected(!model.getViewSchueler());

        int state = model.getPrintState();
        if (state == Model.PRINT_VORNAME) {
            radioPrint1.setSelected(true);
        }
        if (state == Model.PRINT_NACHNAME) {
            radioPrint2.setSelected(true);
        }
        if (state == Model.PRINT_VORNAME_NACHNAME) {
            radioPrint3.setSelected(true);
        }
        if (state == Model.PRINT_NACHNAME_VORNAME) {
            radioPrint4.setSelected(true);
        }

    }

    private void buttonImportCSV() throws UnsupportedEncodingException, FileNotFoundException, IOException {
        // Chooser
        File lastOpenedDirectory = model.getDirectoryLastOpened();
        if (lastOpenedDirectory != null) {
            fileChooserCSV.setCurrentDirectory(lastOpenedDirectory);
        }
        if (fileChooserCSV.showOpenDialog(this) != JFileChooser.APPROVE_OPTION) {
            return;
        }
        File file = fileChooserCSV.getSelectedFile();

        // File --> List
        BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream(file), "UTF8"));
        LinkedList<String> lines = new LinkedList<String>();
        while (br.ready()) {
            lines.addLast(br.readLine());
        }
        br.close();

        String firstLine = lines.pollFirst();
        if (firstLine.split(";").length != 2) {
            JOptionPane.showMessageDialog(this, "Format ist leider keine TXT-Namensliste!");
            return;
        }

        model.namesSetClear(50);

        int j = 0;
        for (int i = 0; i < lines.size(); i++) {
            String[] temp = lines.get(i).split(";");
            if (temp.length == 2) {
                String nachname = temp[0].replaceAll("\"", "");
                String vorname = temp[1].replaceAll("\"", "");
                model.namesSetNachnameAtRow(nachname, j);
                model.namesSetVornameAtRow(vorname, j);
                j++;
            }
        }

    }

}
